//
// Manager.cpp
//

#include "Stdafx.h"
#include "RNSO.h"
#include "GenericObserver.h"
#include "Manager.h"


GenericObjectsIds::GenericObjectsIds()
{
	m_dwNextID = 0;
}


DWORD	GenericObjectsIds::GenerateID(void * pObject)
{
	m_dwNextID++;

	m_adwUsedIDs.Add(m_dwNextID);
	m_apObject.Add(pObject);
	return m_dwNextID;
}

int	GenericObjectsIds::Find(DWORD dwID,void*& pObject)
{
	pObject=NULL;
	for(int i = 0; i < m_adwUsedIDs.GetSize(); i++ )
		if( dwID == m_adwUsedIDs[i] )
		{
			pObject=m_apObject[i];
			return i;
		}

	return -1;
}


void GenericObjectsIds::RemoveObjects(void * pObject)
{
	BOOL bReDo=TRUE;

	while( bReDo )
	{
		bReDo=FALSE;
		for(int i = 0; i < m_apObject.GetSize(); i++ )
		{
			if( m_apObject[i]==pObject )
			{
				m_apObject.RemoveAt(i);
				m_adwUsedIDs.RemoveAt(i);
				bReDo=TRUE;
				break;
			}
		}
	}
}



GenericSubscription::GenericSubscription(	DWORD dwObserverID,void * pObserver,
						DWORD dwSubjectID,void * pSubject)
{
	m_dwObserver=dwObserverID;
	m_dwSubject=dwSubjectID;
	m_pObserver=pObserver;
	m_pSubject=pSubject;
}


GenericObjectsIds	GenericManager::m_aUsedObjectIDs;
CCriticalSection	GenericManager::m_csServer;
CCriticalArray<GenericSubscription*,GenericSubscription*>	GenericManager::m_aSubscriptions;
HANDLE	GenericManager::m_hThread=NULL;
CCriticalArray<ObserverNotification*,ObserverNotification*>	GenericManager::m_aNotifications;
BOOL				GenericManager::m_bContinueToRun;
CEvent				GenericManager::m_evtStopThread;
DWORD				GenericManager::m_ThreadId;
CEvent				GenericManager::m_evtNotification;

DWORD	GenericManager::GenerateObjectID(BOOL& bOk,void * pObject)
{
	DWORD dwID ;
	
	if( bOk=m_csServer.Lock() )
	{
		dwID = m_aUsedObjectIDs.GenerateID(pObject);
		m_csServer.Unlock();
	}
	return dwID;
}


BOOL	GenericManager::AddSubscription(DWORD dwObserverID,DWORD dwSubjectID)
{
	if( ! m_csServer.Lock() )
		return FALSE;

	
	int idxObserver=-1,idxSubject=-1;
	void * pObserver = NULL; 
	void * pSubject  = NULL;

	if( (idxSubject	= m_aUsedObjectIDs.Find(dwSubjectID,pSubject)) !=-1 )
		idxObserver = m_aUsedObjectIDs.Find(dwObserverID,pObserver);

	if( idxObserver != -1 && idxSubject != -1 )
	{
		
		m_aSubscriptions.Add( new GenericSubscription(dwObserverID,pObserver,dwSubjectID,pSubject));
		((CGenericObserver*)pObserver)->IncSub();
		((CGenericObserver*)pSubject)->IncObs();
		
	}
	m_csServer.Unlock();

	return idxObserver != -1 && idxSubject != -1;
}

BOOL	GenericManager::RemoveSubscription(DWORD m_dwObserverID,DWORD dwSubjectID,BOOL bAllSubjects)
{
	if( ! m_csServer.Lock() )
		return FALSE;


	BOOL bReDo = TRUE;
	while( bReDo )
	{
		bReDo=FALSE;
		for( int i = 0; i < m_aSubscriptions.GetSize(); i++ )
		{
			if( m_aSubscriptions[i]->m_dwObserver==m_dwObserverID )
			{
				if( bAllSubjects || m_aSubscriptions[i]->m_dwSubject==dwSubjectID )
				{
					GenericSubscription* pSubscription=m_aSubscriptions[i];
					((CGenericObserver*)pSubscription->m_pObserver)->SubSub();
					((CGenericObserver*)pSubscription->m_pSubject)->SubObs();
					delete pSubscription;
					m_aSubscriptions.RemoveAt(i);
					bReDo = TRUE;
					break;
				}
			}
		}
	}

	m_csServer.Unlock();
	return TRUE;
}


DWORD   _stdcall GenericManager::ThreadProc( LPVOID lpParam ) 
{
#define NBHANDLES	2
	HANDLE handles[NBHANDLES]=	{	(HANDLE) m_evtNotification,
									(HANDLE) m_evtStopThread 
						};

	DWORD dwWait;
	while ( m_bContinueToRun )
	{
		// wait a notification to do
		// send notification
		dwWait = WaitForMultipleObjects(NBHANDLES,handles,FALSE,500);
		if( dwWait == WAIT_FAILED )
			continue;

		if( dwWait >= WAIT_OBJECT_0 && dwWait <= WAIT_OBJECT_0+NBHANDLES-1 )
		{
			if ( dwWait == WAIT_OBJECT_0 + 1)
				break;	// stop the thread..
			else
			{
				// notification..
				while( m_aNotifications.GetSize() && m_bContinueToRun )
				{
					MessageBeep(100);
					m_aNotifications.Lock();
					ObserverNotification * pNotification = m_aNotifications.GetAt(0);
					m_aNotifications.RemoveAt(0);
					m_aNotifications.Unlock();
					DispatchNotification(pNotification);
					delete pNotification;
				}
			}
		}
		else
		{
			// abandonned ..
		}

	}
	return 0;
}


void 	GenericManager::DispatchNotification(ObserverNotification *pNotification)
{
	m_csServer.Lock();

	for(int i = 0; i < m_aSubscriptions.GetSize(); i++)
	{
		if( pNotification->m_dwSenderObject == m_aSubscriptions[i]->m_dwSubject )
		{
			((CGenericObserver*)m_aSubscriptions[i]->m_pObserver)->OnNotify(pNotification->m_dwSenderObject,pNotification->m_nSize,pNotification->m_pBytes);
		}
	}

	m_csServer.Unlock();
}

void	GenericManager::StartManager()
{
	m_bContinueToRun=TRUE;
	m_hThread = ::CreateThread(NULL,0,ThreadProc,NULL,
				0, &m_ThreadId );

}

void	GenericManager::StopManager()
{
	m_bContinueToRun=FALSE;
	m_evtStopThread.SetEvent();

	WaitForSingleObject( m_hThread, 10000 );
	::TerminateThread(m_hThread, 98765321);
}


BOOL	GenericManager::AddNotification(ObserverNotification * pObserverNotification)
{
	m_aNotifications.Lock();
	m_aNotifications.Add( pObserverNotification );
	m_aNotifications.Unlock();
	
	m_evtNotification.SetEvent();
	return TRUE;
}

void	GenericManager::RemoveObject(void * pObject)
{
	m_csServer.Lock();
	

	m_aUsedObjectIDs.RemoveObjects(pObject);
	RemoveSubscriptions(pObject);

	m_csServer.Unlock();
	
}

void	GenericManager::RemoveSubscriptions(void * pObject)
{
	BOOL	bReDo = TRUE;

	while( bReDo )
	{
		bReDo=FALSE;
		for(int i = 0; i < m_aSubscriptions.GetSize(); i++ )
		{
			GenericSubscription* pSubscription = m_aSubscriptions[i];
			if( pSubscription->m_pObserver==pObject ||
				pSubscription->m_pSubject==pObject )
			{
				// Notify the observer that the subject is down
				((CGenericObserver*)pSubscription->m_pObserver)->SubSub();
				((CGenericObserver*)pSubscription->m_pSubject)->SubObs();

				if( pSubscription->m_pSubject==pObject )
				{
					((CGenericObserver*)(pSubscription->m_pObserver)) 
						->
						OnNotifySubjectBroken(pSubscription->m_dwSubject);
				}

				m_aSubscriptions.RemoveAt(i);
				delete pSubscription;
				bReDo= TRUE;
				break;
			}
		}
	}

}

BOOL	GenericManager::IsObserverSubscribedToSubject(DWORD dwObserverID,DWORD dwSubjectID)
{
	BOOL bSubscribed = FALSE;
	m_csServer.Lock();
	for(int i = 0; i < m_aSubscriptions.GetSize(); i++ )
		if( m_aSubscriptions.GetAt(i)->m_dwObserver	== dwObserverID &&
			m_aSubscriptions.GetAt(i)->m_dwSubject	== dwSubjectID )
		{
			bSubscribed = TRUE;
			break;
		}

	m_csServer.Unlock();
	return bSubscribed;
}


BOOL	GenericManager::FindObjectID(BSTR strname,DWORD * dwObjectID)
{
	m_csServer.Lock();
	BOOL bFound = FALSE;
	for(int i = 0; i < m_aUsedObjectIDs.m_apObject.GetSize(); i++ )
	{
		CGenericObserver* pObject=(CGenericObserver*)
											m_aUsedObjectIDs.m_apObject[i];
		if( pObject->m_sname==strname )
		{
			*dwObjectID=pObject->m_dwObserverID;
			bFound = TRUE;
			break;
		}
	}
	m_csServer.Unlock();
	return bFound;
	
}

int		GenericManager::FindName(CComBSTR bstr)
{
	int iCount = 0;
	m_csServer.Lock();
	for(int i = 0; i < m_aUsedObjectIDs.m_apObject.GetSize(); i++ )
	{
		CGenericObserver* pObject=(CGenericObserver*)
											m_aUsedObjectIDs.m_apObject[i];
		if( pObject->m_sname==bstr )
			iCount ++;
	}
	m_csServer.Unlock();
	return iCount;
}
